﻿using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Operators;

namespace HeuristicLab.Problems.PTSP {
  [Item("Exhaustive25MoveGenerator", "Generates all possible inversion and shift moves (2.5-opt) from a given permutation.")]
  [StorableClass]
  class Exhaustive25MoveGenerator : TwoPointFiveMoveGenerator, IExhaustiveMoveGenerator {
    public override bool CanChangeName {
      get { return false; }
    }

    [StorableConstructor]
    protected Exhaustive25MoveGenerator(bool deserializing) : base(deserializing) { }
    protected Exhaustive25MoveGenerator(Exhaustive25MoveGenerator original, Cloner cloner) : base(original, cloner) { }
    public Exhaustive25MoveGenerator() 
      : base() {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new Exhaustive25MoveGenerator(this, cloner);
    }
    
    public static TwoPointFiveMove[] Apply(Permutation permutation) {
      return Generate(permutation).ToArray();
    }

    public static IEnumerable<TwoPointFiveMove> Generate(Permutation permutation) {
      int length = permutation.Length;
      if (length == 1) throw new ArgumentException("Exhaustive25MoveGenerator: There cannot be a 2.5 move given a permutation of length 1.", "permutation");
      int totalMoves = (length) * (length - 1) / 2;

      if (permutation.PermutationType == PermutationTypes.RelativeUndirected) {
        if (totalMoves - 3 > 0) {
          for (int i = 0; i < length - 1; i++) {
            for (int j = i + 1; j < length; j++) {
              // doesn't make sense to inverse the whole permutation or the whole but one in case of relative undirected permutations
              if (j - i >= length - 2) continue;
              yield return new TwoPointFiveMove(i, j,true);
              yield return new TwoPointFiveMove(i, j, false);
            }
          }
        } else { // when length is 3 or less, there's actually no difference, but for the sake of not crashing the algorithm create a dummy move
          yield return new TwoPointFiveMove(0, 1,true);
        }
      } else {
        for (int i = 0; i < length - 1; i++)
          for (int j = i + 1; j < length; j++) {
            yield return new TwoPointFiveMove(i, j,true);
            yield return new TwoPointFiveMove(i, j, false);
          }
      }
    }

    protected override TwoPointFiveMove[] GenerateMoves(Permutation permutation) {
      return Apply(permutation);
    }
  }
}
